# add_sofa_library: Adds a SOFA library to the project.
#
#   add_sofa_library(<name> [NO_PREFIX_UNDERSCORE | NO_PREFIX] [shared_symbol])
#
# Adds a library target to the project using the following SOFA conventions:
#
#   (1) The library name is prefixed by "sofa_". Set the NO_PREFIX_UNDERSCORE
#       flag to omit the underscore of the prefix. To omit the prefix
#       completely specify the NO_PREFIX flag instead.
#
#   (2) While building the library itself SOFA_BUID_<NAME> is defined.
#       This is especially useful for specifying the right storage-class
#       attributes (i.e. dllexport or dllimport) during the build and
#       actual usage of the library. <NAME> can be overwritten by
#       specifying shared_symbol.
#
# The function checks for two variables that contain the sources and target
# link libraries. Their name must be <NAME>_SOURCES and <NAME>_LIBS. The
# latter is optional.
#
# Example:
#
#   set(MISC_ENGINE_SOURCES engine/Distances.cpp initMiscEngine.cpp)
#   set(MISC_ENGINE_LIBS sofa_non_uniform_fem)
#   add_sofa_library(misc_engine)

function(add_sofa_library)
  string(TOUPPER ${ARGV0} uppercase_name)
  if(ARGV1)
    string(COMPARE EQUAL NO_PREFIX_UNDERSCORE ${ARGV1} no_prefix_underscore)
    if(${no_prefix_underscore})
      set(prefixed_name sofa${ARGV0})
    else()
      string(COMPARE EQUAL NO_PREFIX ${ARGV1} no_prefix)
      if(${no_prefix})
        set(prefixed_name ${ARGV0})
      else()
        set(shared_symbol SOFA_BUILD_${ARGV1})
      endif()
    endif()
  endif()
  if(NOT prefixed_name)
    set(prefixed_name sofa_${ARGV0})
  endif()
  if(NOT shared_symbol)
    if(ARGV2)
      set(shared_symbol SOFA_BUILD_${ARGV2})
    else()
      set(shared_symbol SOFA_BUILD_${uppercase_name})
    endif()
  endif()
  add_library(${prefixed_name} ${${uppercase_name}_SOURCES})
  if(${uppercase_name}_LIBS)
    target_link_libraries(${prefixed_name} ${${uppercase_name}_LIBS})
  endif()
  set_target_properties(${prefixed_name} PROPERTIES
    DEFINE_SYMBOL ${shared_symbol}
  )
  set(SOFA_LIBRARIES "${SOFA_LIBRARIES}${prefixed_name} " CACHE STRING "" FORCE)
endfunction()